(*
 * SNU 4190.310 Programming Languages (2011 Spring)
 *
 * K- Interpreter I
 *)

(* Location *)
module type LOC =
  sig
    (* Base Address *)
    module Base :
      sig
        (* type of bass address *)
        type t

        (* NULL *)
        val null : t

        (* successor *)
        val succ : t -> t

        (* equal b1 b2 tests equality of b1 and b2 *)
        val equal : t -> t -> bool
      end

    (* type of location *)
    type t = Base.t * tyoffset
    and tyoffset = int

    exception Not_allowed

    (* loc_of_base b returns the first (offset 0) location of b *)
    val loc_of_base : Base.t -> t

    (* base l returns the base address of l. *)
    val base : t -> Base.t

    (* offset l return the offset of l. *)
    val offset : t -> int

    (* addi l i adds i to the offset of l. *)
    val addi : t -> int -> t

    (* subi l i subtracts i from the offset of l. *)
    val subi : t -> int -> t

    (* sub x y returns (offset x) - (offset y), or raises Not_allowed if they
       have different base addresses. *)
    val sub : t -> t -> tyoffset

    (* equal l r tests equality of l and r, or raises Not_allowed if they have
       different base addresses. *)
    val equal : t -> t -> bool

    (* less l r returns (offset l) < (offset r), or raises Not_allowed if they
       have different base addresses. *)
    val less : t -> t -> bool
  end

module Loc : LOC =
  struct
    module Base =
      struct
        type t = int
        let null = 0
        let succ b = b + 1
        let equal l r = (l = r)
      end

    type t = Base.t * tyoffset
    and tyoffset = int

    exception Not_allowed

    let loc_of_base base = (base, 0)
    let base (base, _) = base
    let offset (_, offset) = offset
    let addi (base, offset) n = (base, offset + n)
    let subi (base, offset) n = (base, offset - n)
    let sub (base, offset) (base', offset') =
        if Base.equal base base' then offset - offset' else raise Not_allowed
    let equal (base, offset) (base', offset') =
        (Base.equal base base') && (offset = offset')
    let less (base, offset) (base', offset') =
        if Base.equal base base' then offset < offset' else raise Not_allowed
  end

(*
 * Memory
 *)
module type MEM =
  sig
    (* type of memory from Loc.t to 'a *)
    type 'a t

    exception Not_allocated

    (* empty memory *)
    val empty : 'a t

    (* allocate m s returns an Loc.t to a fresh (uninitialized) buffer of size
       s and a memory containing the same bindings as m, plus a binding of
       the Loc.t to the first location of allocated buffer. *)
    val allocate : 'a t -> int -> Loc.t * 'a t

	(* "alloc m" equals to "allocate m 1" *)
	val alloc : 'a t -> Loc.t * 'a t

    (* store m l v returns a memory containing the same bindings as m, with l
       bound to v, or raises Not_allocated if m doesn't contain a binding for
       l. *)
    val store : 'a t -> Loc.t -> 'a -> 'a t

    (* load m l returns the current binding of l in m, or raises Not_found if
       no such binding exists. *)
    val load : 'a t -> Loc.t -> 'a

    (* is_allocated m l returns true if m contains a binding for l, and false
       otherwise. *)
    val is_allocated : 'a t -> Loc.t -> bool
  end

module Mem : MEM =
  struct
    module Map = Map.Make (struct type t = Loc.t let compare = compare end)
    type 'a t = Loc.Base.t * ('a option Map.t)

    exception Not_allocated
	exception Not_initialized

    let empty = (Loc.Base.null, Map.empty)

    let allocate (base, mem) s =
    	let rec range f t = if f < t then f :: (range (f + 1) t) else [] in

        let base' = Loc.Base.succ base in
        let l = Loc.loc_of_base base' in
        	(l, (base', List.fold_left
                        (fun m o -> Map.add (Loc.addi l o) None m)
                        mem
                        (range 0 s)))

	let alloc m = allocate m 1

    let is_allocated (_, mem) l = Map.mem l mem

    let store m l v =
		let (base, mem) = m in
        if is_allocated m l then (base, Map.add l (Some v) mem)
        else raise Not_allocated

    let load (_, mem) l =
		try
	        let content = (Map.find l mem) in 
				match content with
				  None -> raise Not_initialized
				| Some v -> v
		with Not_found -> raise Not_allocated
  end

(*
 * Environment
 *)
module type ENV =
  sig
    (* type of environment from 'a to 'b *)
    type 'a t

    exception Not_bound

    (* empty environment *)
    val empty : 'a t

    (* bind e x l returns an environment containing the same bindings as e,
       plus a binding of x to l. If x was already bound in e, its previous
       binding disappears. *)
    val bind : 'a t -> string -> 'a -> 'a t

    (* lookup e x returns the current binding of x in e, or raises Not_bound if
       no such binding exists. *)
    val lookup : 'a t -> string -> 'a

    (* is_bound e x returns true if e contains a binding for x, and false
       otherwise. *)
    val is_bound : 'a t -> string -> bool
  end

module Env : ENV =
  struct
  	module Map = Map.Make (struct type t = string let compare = compare end)
    type 'a t = 'a Map.t
   
	exception Not_bound

    let empty = Map.empty

    let bind e x a = Map.add x a e

    let lookup e x = try (Map.find x e) with Not_found -> raise Not_bound

    let is_bound e x = Map.mem x e
  end

(*
 * K- Interpreter
 *)
module type KMINUS =
  sig
    exception Error of string
    type id = string
    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | SEQ of exp * exp                   (* sequence *)
      | IF of exp * exp * exp              (* if-then-else *)
      | WHILE of exp * exp                 (* while loop *)
      | LETV of id * exp * exp             (* variable binding *)
      | LETF of id * id list * exp * exp   (* procedure binding *)
      | CALLV of id * exp list             (* call by value *)
      | CALLR of id * id list              (* call by referenece *)
      | RECORD of (id * exp) list          (* record *)
      | FIELD of exp * id                  (* record field access *)
      | ASSIGN of id * exp                 (* assgin to variable *)
      | ASSIGNFIELD of exp * id * exp      (* assign to recoed field *)
      | MALLOC of exp                      (* malloc *)
      | AMPER of id                        (* address of id *)
      | AMPERFIELD of exp * id             (* address of record field *)
      | STAR of exp                        (* dereference *)
      | ASSIGNSTAR of exp * exp            (* assign to locatio *)
      | READ of id
      | WRITE of exp

    type program = exp
    type memory
    type env
    type value

	val emptyMemory : memory
    val emptyEnv : env
    val run : memory * env * program -> value
  end

module K : KMINUS =
  struct
    exception Error of string

    type id = string

    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | SEQ of exp * exp
      | IF of exp * exp * exp
      | WHILE of exp * exp
      | LETV of id * exp * exp
      | LETF of id * id list * exp * exp
      | CALLV of id * exp list
      | CALLR of id * id list
      | RECORD of (id * exp) list
      | FIELD of exp * id
      | ASSIGN of id * exp
      | ASSIGNFIELD of exp * id * exp
      | MALLOC of exp
      | AMPER of id
      | AMPERFIELD of exp * id
      | STAR of exp
      | ASSIGNSTAR of exp * exp
      | READ of id
      | WRITE of exp
    type program = exp

    type value =
        Num of int
      | Bool of bool
      | Location of Loc.t
	  | Record (* IMPLEMENT HERE *)
      | Unit

    (* memory *)
    type memory = value Mem.t

    let emptyMemory = Mem.empty

    (* environment *)
    type env = env_entry Env.t
    and  env_entry = Addr of Loc.t | Proc of id list * exp * env

    let emptyEnv = Env.empty

    let env_loc e x =
		try
        (match Env.lookup e x with
           Addr l -> l
         | Proc _ -> raise (Error "not allowed")) 
		with Env.Not_bound -> raise (Error "not bound")

    let env_proc e f = 
		try
        (match Env.lookup e f with
           Addr _ -> raise (Error "not allowed") 
         | Proc (idlist, exp, env) -> (idlist, exp, env)) 
		with Env.Not_bound -> raise (Error "not bound")

	(* IMPLEMENT FUNCTIONS HERE IF YOU THINK YOU NEED TO *)

	let rec eval : memory * env * exp -> value * memory =
		fun (mem, env, e) -> match e with

		(* IMPLEMENT HERE *)

		| _ -> raise (Error "not implemented")

	(* OR HERE, IF YOU THINK YOU NEED TO *)

	let run (mem, env, pgm) =
		let (v,_) = eval (mem, env, pgm) in
		v

  end
